template_config = '''
# Sample config.yaml file to generate a random program
# using aapg (Automated Assembly Program Generator)

# Each section commands a behaviour of aapg
# and inline comments in each section will explain the
# usage
# ---------------------------------------------------------------------------------
# Privlege mode that instruction are executed in
# Options:
#       mode: m/s/u
# Note: If the privlege mode is either s or u, then the test_entry_macro
# will be defined accordingly
# ---------------------------------------------------------------------------------
priv-mode:
  mode: m
# ---------------------------------------------------------------------------------
# General directives to aapg
# Options:
#       total_instructions: Approximate number of instructions to be generated
#                           by aapg. Actual may vary.
#       regs_not_use:       Comma separated list of RISC-V registers to not use for
#                           reading/writing in the random generated instructions
#                           
# ---------------------------------------------------------------------------------
general:
  total_instructions: 1000
  regs_not_use: x1,x2
  custom_trap_handler: true
  code_start_address: 0x80000000
  default_program_exit: true
  delegation: 0xfff

# ---------------------------------------------------------------------------------
# Distribution of instructions according to ISA extensions
# Specify the relative frequency of each set 
# E.g. : A relative frequency of 1 each means each instruction will be 
# generated equal number of times in the total instructions. Specify 0 to disable.
# ---------------------------------------------------------------------------------
isa-instruction-distribution:
  rel_sys: 0
  rel_sys.csr: 0
  rel_rv32i.ctrl: 0
  rel_rv32i.compute: 1
  rel_rv32i.data: 0
  rel_rv32i.fence: 0
  rel_rv64i.compute: 0
  rel_rv64i.data: 0
  rel_rv32m: 0
  rel_rv64m: 0
  rel_rv32a: 0
  rel_rv64a: 0
  rel_rv32f: 0
  rel_rv64f: 0
  rel_rv32d: 0
  rel_rv64d: 0

# Compressed instructions

  rel_rvc.ctrl: 0
  rel_rvc.compute: 0
  rel_rvc.sp: 0
  rel_rvc.data: 0
  rel_rvc.fdata: 0

  rel_rv32c.compute: 0
  rel_rv32c.ctrl: 0
  rel_rv32c.fdata: 0

  rel_rv64c.compute: 0
  rel_rv64c.data: 0

float-rounding:
  rne: 10
  rtz: 10
  rdn: 10
  rup: 10
  rmm: 10

branch-control:
  backward-probability: 0.5
  block-size: 7

# ---------------------------------------------------------------------------------
# Recursion options
# Options:
#       recursion_enable:   Generate the template for recursion or not
#       recursion_depth:    Number of times the recursive function is called
# ---------------------------------------------------------------------------------
recursion-options:
  recursion-enable: false
  recursion-depth: 10
  recursion-calls: 5

# ---------------------------------------------------------------------------------
# Data access sections
# Specify which regions of memory will be accessed by the random program
# Options:
#       enable:         Force all memory accesses instructions to only load/store
#                       to the specified list of address sections
#                       
# Section Template: Specify legal access zones using the following template
#       section_name:  section_low,section_high (HEX)     
# ---------------------------------------------------------------------------------
access-sections:
  begin_signature: 0x80091000,0x80095000,rw


# ---------------------------------------------------------------------------------
# CSR sections
# Specify which CSRs will be accessed by the random program
# Options:              
#       sections:
#                 start-value1:end-value1, value2, start-value3:end-value3 (HEX)
#                 (Any Combination)
# ---------------------------------------------------------------------------------
csr-sections:
  sections: 0x0100:0xdff, 0x325, 0x500:0xfff


# ---------------------------------------------------------------------------------
# User template sections
# Allows users to specify call to a custom function with number of times to call
# User should ensure that function does not modify 
#                       
# Section Template: Specify user template function calls with the number of times
#       function_name: '{number_of_times:"function_body"}'
# ---------------------------------------------------------------------------------
user-functions:
  func1: '{0:"add x0,x0,x0"}'

# ---------------------------------------------------------------------------------
# Switching Privledge modes in AAPG
#       switching_modes:  true/false (Do not provide any user defined functions when
#                         shifting mdoes is true)
#       num_switches:     # of times privlege modes has to shift (This is randomised 
#                         and shifting may result in same mode)
# ---------------------------------------------------------------------------------
switch-priv-modes:
  switch_modes: false
  num_switches: 0


# ---------------------------------------------------------------------------------
# Instruction Cache and Data-Cache Thrashing
# ---------------------------------------------------------------------------------
i-cache:
  num_calls: 0
  num_bytes_per_block: 16
  num_blocks: 8
  num_cycles: 10

d-cache:
  num_calls: 0
  num_bytes_per_block: 16
  num_blocks: 8
  num_cycles: 10

# ---------------------------------------------------------------------------------
# Exception generation
# ---------------------------------------------------------------------------------
exception-generation:
  ecause00: 0
  ecause01: 0
  ecause02: 0
  ecause03: 0
  ecause04: 0
  ecause05: 0
  ecause06: 0
  ecause07: 0
  ecause08: 0
  ecause09: 0
  ecause10: 0
  ecause11: 0
  ecause12: 0
  ecause13: 0
  ecause14: 0

# ---------------------------------------------------------------------------------
# Data Hazards
# ---------------------------------------------------------------------------------
data-hazards:
  raw_prob: 0.5
  war_prob: 0.5
  waw_prob: 0.5
  num_regs_lookbehind: 3

program-macro:
  pre_program_macro: "add x0,x0,x0"
  post_program_macro: "add x0,x0,x0"
  pre_branch_macro: "add x0,x0,x0"
  post_branch_macro: "add x0,x0,x0"
  ecause00: "random"
  ecause01: "random"
  ecause02: "random"
  ecause03: "random"
  ecause04: "random"
  ecause05: "random"
  ecause06: "random"
  ecause07: "random"
  ecause08: "random"
  ecause09: "random"
  ecause10: "random"
  ecause11: "random"
  ecause12: "random"
  ecause13: "random"
  ecause14: "random"

# ---------------------------------------------------------------------------------
# Self Checking 
# If self_checking flag is enabled during the gen command, the below section is used
# Options:              
#       rate: The interval at which checksums are added to the test. 
#             If rate = 10, checksums will be added every 10 instructions
# ---------------------------------------------------------------------------------

self-checking:
  rate: 100
  test_pass_macro: "la      sp, begin_signature;\n addi    sp, sp, 2*REGBYTES;\n li      t1, 0xfffff;\n SREG    t1, 0*REGBYTES(sp)"
  test_fail_macro: "add x0,x0,x0"
'''



"""
    Module that contains random program generators

    Each program generator is a state machine that
    generates one instruction at a time. Each object
    creates a generator which can be iterated over
"""
from six.moves import queue
from six.moves import configparser
import logging

import aapg.utils
import aapg.isa_funcs
import aapg.args_generator
import aapg.asm_templates

import random
import re
import os
import sys

""" Setup the logger """
logger = logging.getLogger(__name__)

class BasicGenerator(object):
    """ Basic Generator to generate random instructions """

    def __init__(self, args, arch, seed, no_use_regs, self_checking):
        logger.debug("Created instance of BasicGenerator")

        # Instantiate local variables
        self.q = queue.Queue()
        self.total_instructions = int(args.get('general', 'total_instructions'))
        self.ref_total_instructions = self.total_instructions
        self.inst_dist = None
        self.inst_dist_nobranch = None
        self.inst_dist_backup = None
        self.inst_dist_nobranch_backup = None
        self.regfile = {}
        self.local_regfile = None
        self.instructions_togen = 0
        self.args = args
        self.end_sections = False
        self.recursion_enabled = args.getboolean('recursion-options', 'recursion-enable')
        self.arch = arch
        self.seed = seed
        self.current_access_section = None
        self.end = False
        self.data_hazards = args.items('data-hazards')
        self.curr_inst_nolabel = False
        self.branch_use_reg = None
        self.no_use_regs = no_use_regs
        self.reg_ignore = None
        self.rec_use_reg1 = None
        self.rec_use_reg2 = None
        self.skip_user_flag = False
        self.csr_sections  = args.get('csr-sections','sections')
        self.branch_block_size = args.get('branch-control','block-size')
        self.delegation_boolean = False
        self.self_checking = self_checking
        try:
            self.del_input = int(args.get('general','delegation'))
            if self.del_input != 0:
                self.delegation_boolean = True
        except:
            logger.warn('Delegation Value not provided, Delegation Not Enabled')
            self.del_input = 0xfff
            self.delegation_boolean = False

        # Create the data_access sections
        access_sections = args.items('access-sections')
        self.access_sections = []
        for item in access_sections:
            bounds = item[1].split(',')
            lower_bound = int(bounds[0], 16)
            upper_bound = int(bounds[1], 16)
            read_write = bounds[2]
            label = item[0]
            self.access_sections.append((lower_bound, upper_bound, read_write, label))

        # Seeding the PRNG generator
        random.seed(self.seed)
        aapg.args_generator.set_seed_args_gen(self.seed)
        aapg.isa_funcs.set_seed_isa_funcs(self.seed)

        # Read the instruction distribution
        try:
            self.compute_instruction_distribution(args.items('isa-instruction-distribution'))
            logger.debug("Instruction distribution received")
            for k in self.inst_dist:
                logger.debug("{0} - {1}".format(k, self.inst_dist[k]))
        except configparser.NoSectionError as e:
            logger.error("Instruction distribution not specified.")
            logger.error("Check if your config file has the [isa-instruction-distribution] section")
            sys.exit(1)

        # Setup the register file
        self.init_regfile(args.get('general', 'regs_not_use'))

        # Add a call to the pre-program macro
        logger.info("Adding pre-program macro")

        # Create Pre-lude
        logger.info("Creating Prelude")
        self.add_prelude(args)

        # Set SP to a legal range
        self.add_memory_instruction(random_gen = False)

        # Log debug messages
        logger.debug("Total instructions: {0}".format(self.total_instructions))
        logger.debug("Total instructions to generate: {0}".format(self.instructions_togen))

        # Branch control
        self.insts_since_last_branch = 0
        if self.branch_ext_selected(self.inst_dist):
            self.branch_bwd_prob = args.get('branch-control', 'backward-probability')
            logger.info("Branch instructions present")

        # Adding this to act as counter for branches
        if self.branch_use_reg!=None:
            self.q.put(('instruction', ['li', 'x{branch_use_reg}'.format(branch_use_reg=self.branch_use_reg), '10']))
        else:
            self.q.put(('instruction', ['li', 'x31', '10']))

        if self.self_checking:
            self.q.put(('instruction_nolabel', ('la', 't0', 'register_swap')))
            self.q.put(('instruction_nolabel', ('csrw', 'mscratch', 't0')))
            self.q.put(('instruction', ['li', 'x5', '0']))

        self.total_instructions += 1

        # Setup the user function call 
        try:
            sci = int(args.get('self-checking','rate'))
        except:
            sci = 100


        self.user_calls_dict = {'user-functions' : args.items('user-functions')}
        self.user_calls_dict['i_cache_thrash'] = ('f', args.getint('i-cache', 'num_calls'))
        if args.getboolean('switch-priv-modes', 'switch_modes') and self.self_checking != True:
            self.user_calls_dict['switchmodes'] = ('m', args.getint('switch-priv-modes', 'num_switches'))

        self.track_chsum = 0
        self.count_chsum = 0
        self.self_checking_interval = sci
        self.total_chsum = args.getint('general','total_instructions')/sci
        
        ecause_filtered = list(filter(lambda x: int(x[1]) > 0, args.items('exception-generation')))

        for ecause in ecause_filtered:
            self.user_calls_dict[ecause[0]] = ('m', int(ecause[1]))

        keys = ' '.join(self.user_calls_dict.keys())

        logger.info('User functions received {}'.format(keys))
        new_dict = {} 

        for key, value in self.user_calls_dict.items():
            if key.startswith('ecause'):
                (mode,number) = value
                for i in range(number):
                    new_key = key+'_{:05d}'.format(i)
                    new_value = ('m',1)
                    new_dict.update({new_key:new_value})
            elif key.startswith('user-functions'):
                for (key_in,value_in) in value:
                    new_key = key_in
                    in_dict = eval(value_in)
                    for k,v in in_dict.items():
                        new_value = ('f',int(k))
                    new_dict.update({new_key:new_value})
            else:
                new_dict.update({key:value})

        self.user_calls_dict = new_dict

        # Add recursion call
        if self.recursion_enabled:
            self.user_calls_dict['.recurse_init'] = ('f', int(args.get(
                    'recursion-options', 'recursion-calls')))

    def __iter__(self):
        return self

    def __next__(self):
        if self.total_instructions != 0:
            self.generate_next_instruction()

        if self.q.empty():
            if not self.end:
                self.q.put(('instruction_nolabel', ('post_program_macro', )))

                if self.args.getboolean('general', 'default_program_exit'):
                    self.q.put(('pseudo', ['j', 'write_tohost']))
                    self.total_instructions += 1
                self.total_instructions += 1
                self.end = True
            elif self.recursion_enabled:
                self.add_recursion_sections(self.args.getint('recursion-options', 'recursion-depth'))
                self.recursion_enabled = False
            else:
                raise StopIteration('Instructions over')

        return self.q.get()
    
    def next(self):
        """Python 2 compat"""
        return self.__next__()

    def generate_next_instruction(self):
        # Check if total number of required instructions have been generated
        if self.total_instructions == 0:
            logger.info("Total number of instructions required generated")
            return

        if self.self_checking:
                self.track_chsum = self.track_chsum + 1

        if self.self_checking:
            if self.track_chsum >= self.self_checking_interval:
                if self.count_chsum < self.total_chsum + 1:
                    self.track_chsum = 0
                    self.count_chsum = self.count_chsum + 1
                    self.q.put(('instruction', ('call', 'write_chsum')))

        # Randomly add a user-defined call
        temp_dict = self.user_calls_dict
        self.user_calls_dict = {x:temp_dict[x] for x in temp_dict if temp_dict[x][1] > 0}
        user_defined_total_calls = sum(map(lambda x: x[1], self.user_calls_dict.values()))
        logger.debug('User calls left {}'.format(user_defined_total_calls))



        if user_defined_total_calls > 0 and random.uniform(0,1) < (user_defined_total_calls/self.instructions_togen) and self.skip_user_flag==False:  #0.84 for 80%
            
            logger.debug('Adding a user call')

            # Select the user call
            user_call = random.choice(list(self.user_calls_dict.keys()))

            # Check if function call or macro
            if self.user_calls_dict[user_call][0] == 'f':
                self.q.put(('instruction', ('call', user_call)))
            else:
                self.q.put(('instruction', (user_call, )))
            self.user_calls_dict[user_call] = (self.user_calls_dict[user_call][0], self.user_calls_dict[user_call][1] - 1)

            self.total_instructions -= 1

            # Check if only user defined calls are left (To avoid grouping at the bottom)
            if self.total_instructions <= user_defined_total_calls or self.instructions_togen == 0:
                logger.warn("Skipping adding user function to test, to maintain even distribution")
                usrfprt = str(self.user_calls_dict)
                logger.warn("Functions Skipped: "+usrfprt)
                self.skip_user_flag = True
                self.total_instructions = 0
                return

            return
            

        # Select a random instruction
        if self.instructions_togen > 0:
            # To keep track if stuck in while loop
            reset_threshold = 0
            next_inst_found = False
            while not next_inst_found:

                # If we have crossed a certain threshold from the
                # previous branch instruction, generate a branch
                # instruction possibly, otherwise omit the key
                if self.branch_ext_selected(self.inst_dist):
                    if self.insts_since_last_branch > self.min_insts_threshold:
                        isa_ext = random.choice(self.branch_exts(self.inst_dist))
                    else:
                        isa_ext = random.choice(list(self.inst_dist_nobranch.keys()))
                else:
                    isa_ext = random.choice(list(self.inst_dist.keys()))

                if self.inst_dist[isa_ext] > 0:
                    next_inst = aapg.isa_funcs.get_random_inst_from_set(isa_ext)
                    self.inst_dist[isa_ext] -= 1

                    if not self.is_branch_ext(isa_ext):
                            self.insts_since_last_branch += 1

                    next_inst_found = True

                reset_threshold = reset_threshold + 1
                # Reset instruction distribution if stuck in while loop
                if reset_threshold > 1000:
                    logger.warn("Error in selecting instruction, Resetting distribution")
                    self.inst_dist = self.inst_dist_backup.copy()
                    self.inst_dist_nobranch = self.inst_dist_nobranch_backup.copy()

            # With a small probablity, write to fcsr register    
            if next_inst[0] in aapg.isa_funcs.float_insts and self.total_instructions > 2 and self.instructions_togen > 2:
                fcsr_prob = random.uniform(0,1)
                if fcsr_prob<0.05:
                    selector = [0,1,2,3,4,5,6,7]
                    choice = random.choice(selector)
                    self.q.put(('instruction', ('fsrmi', 'x0', '{}'.format(str(choice)))))

            # if memory_insts, randomly displace sp
            if next_inst[0] in aapg.isa_funcs.memory_insts and len(self.access_sections) > 0:
                self.add_memory_instruction()
                
                # Choose the offset register
                off_reg = 'sp'

                next_inst = tuple([next_inst[0], next_inst[1], off_reg , next_inst[3]])

                # check data section and replace lw with sw
                if next_inst[0][0] != 'c':
                    if self.current_access_section[2] == 'w':
                        if 'l' in next_inst[0]:
                            next_inst_op = next_inst[0].replace('l', 's', 1)
                            next_inst = (next_inst_op,) + next_inst[1:]
                            logger.debug("Load to write section. Rewriting")
                    if self.current_access_section[2] == 'r':
                        if 's' in next_inst[0]:
                            next_inst_op = next_inst[0].replace('s', 'l', 1)
                            next_inst = (next_inst_op,) + next_inst[1:]
                            logger.debug("Store to read section. Rewriting")
                elif next_inst[0][0] == 'c':
                    if self.current_access_section[2] == 'w':
                        if 'l' in next_inst[0]:
                            next_inst_op = next_inst[0].replace('l', 's', 1)
                            next_inst = (next_inst_op,) + next_inst[1:]
                            logger.debug("Load to write section. Rewriting")
                    if self.current_access_section[2] == 'r':
                        if 's' in next_inst[0] and 'l' not in next_inst[0]:
                            next_inst_op = next_inst[0].replace('s', 'l', 1)
                            next_inst = (next_inst_op,) + next_inst[1:]
                            logger.debug("Store to read section. Rewriting")


            # if control instruction
            if self.is_branch_ext(isa_ext):
                (next_insts,self.reg_ignore) = aapg.args_generator.gen_branch_args(
                        next_inst,
                        self.regfile,
                        self.arch,
                        insts_total = self.ref_total_instructions,
                        insts_left = self.instructions_togen,
                        insts_since = self.insts_since_last_branch,
                        bwd_prob = self.branch_bwd_prob,
                        branch_use_reg = self.branch_use_reg
                )
                
                extra_instructions = []
                self.local_regfile = self.regfile.copy()

                if self.reg_ignore != None:
                    self.reg_ignore = ('x',self.reg_ignore)
                    if self.reg_ignore in self.local_regfile.keys():
                        del self.local_regfile[self.reg_ignore]

                to_temp_remove = [5,6,10,11,12,13,30]
                for i in to_temp_remove:
                    rem = ('x',i)
                    if rem in self.local_regfile.keys():
                        del self.local_regfile[rem]

                no_pre_insts = int(self.branch_block_size) * 2
                comp_string_begin = ['beq','bne','blt','bge','jal','jalr','pre_branch_macro']
                comp_string_end = ['beq','bne','blt','bge','jal','jalr','post_branch_macro']

                
                begin = 0
                if self.reg_ignore!=None:
                    for start in range(len(next_insts)):
                        if next_insts[start][0] == 'la':
                            if next_insts[start][2][0] == 'f':
                                begin = 0
                                break
                        if next_insts[start][0] in comp_string_begin:
                            begin = start
                            break

                end = len(next_insts)
                if self.reg_ignore!=None:
                    for i in range(len(next_insts)-1,-1,-1):
                        if "1:" in next_insts[i][0]:
                            end = len(next_insts)
                            break
                        if next_insts[i][0] in comp_string_end:
                            end = i+1
                            break



                if begin!=0:
                    for i in range(begin):
                        self.q.put(('instruction_nolabel',next_insts[i]))
                for dummy in range(no_pre_insts):

                    isa_ext = random.choice(list(self.inst_dist_nobranch.keys()))
                    next_inst = aapg.isa_funcs.get_random_inst_from_set(isa_ext)

                    
                    if next_inst[0] in aapg.isa_funcs.memory_insts and len(self.access_sections) > 0:
                        self.add_memory_instruction()
                        
                        # Choose the offset register
                        off_reg = 'sp'

                        next_inst = tuple([next_inst[0], next_inst[1], off_reg , next_inst[3]])

                        # check data section and replace lw with sw
                        if next_inst[0][0] != 'c':
                            if self.current_access_section[2] == 'w':
                                if 'l' in next_inst[0]:
                                    next_inst_op = next_inst[0].replace('l', 's', 1)
                                    next_inst = (next_inst_op,) + next_inst[1:]
                                    logger.debug("Load to write section. Rewriting")
                            if self.current_access_section[2] == 'r':
                                if 's' in next_inst[0]:
                                    next_inst_op = next_inst[0].replace('s', 'l', 1)
                                    next_inst = (next_inst_op,) + next_inst[1:]
                                    logger.debug("Store to read section. Rewriting")
                        elif next_inst[0][0] == 'c':
                            if self.current_access_section[2] == 'w':
                                if 'l' in next_inst[0]:
                                    next_inst_op = next_inst[0].replace('l', 's', 1)
                                    next_inst = (next_inst_op,) + next_inst[1:]
                                    logger.debug("Load to write section. Rewriting")
                            if self.current_access_section[2] == 'r':
                                if 's' in next_inst[0] and 'l' not in next_inst[0]:
                                    next_inst_op = next_inst[0].replace('s', 'l', 1)
                                    next_inst = (next_inst_op,) + next_inst[1:]
                                    logger.debug("Store to read section. Rewriting")


                    flag = False
                    if isa_ext == 'rv32a' or isa_ext == 'rv64a':
                        #self.instructions_togen -= 1
                        #self.total_instructions -= 1
                        next_inst_with_args = aapg.args_generator.gen_atomic_args(
                            next_inst,
                            self.local_regfile,
                            self.arch,
                            self.reg_ignore,
                            data_hazards = self.data_hazards
                            )
                        flag = True

                    if flag == False:
                        next_inst_with_args = aapg.args_generator.gen_args(
                            next_inst,
                            self.local_regfile,
                            self.arch,
                            self.reg_ignore,
                            self.csr_sections,
                            total = self.ref_total_instructions,
                            current = self.instructions_togen,
                            data_hazards = self.data_hazards)

                    #Add instruction to branch block
                    extra_instructions.append(('instruction_nolabel', next_inst_with_args))

                #Split the instructions to be before and after the branch instruction
                before = extra_instructions[:len(extra_instructions)//2]
                after = extra_instructions[len(extra_instructions)//2:]

                for i in range(len(before)):
                    self.q.put(before[i])

                temp_list = next_insts[begin:end]
                if end != len(next_insts):
                    temp_list.append(next_insts[-1])
                
                self.q.put(('branch', temp_list))

                for i in range(len(after)):
                    self.q.put(after[i])
                
                if end != len(next_insts):
                    for i in range(end,len(next_insts)-1):
                        self.q.put(('instruction_nolabel', next_insts[i]))

                self.insts_since_last_branch = 0
                self.instructions_togen -= 1
                self.total_instructions -= 1
                self.reg_ignore = None
                self.local_regfile = None
                
                return

            # if atomic instruction
            if isa_ext == 'rv32a' or isa_ext == 'rv64a':
                self.instructions_togen -= 1
                self.total_instructions -= 1
                next_inst_with_args = aapg.args_generator.gen_atomic_args(
                    next_inst,
                    self.regfile,
                    self.arch,
                    self.reg_ignore,
                    data_hazards = self.data_hazards
                    )
                self.q.put(('instruction', next_inst_with_args))

                return

            # Create args for next instruction
            next_inst_with_args = aapg.args_generator.gen_args(
                    next_inst,
                    self.regfile,
                    self.arch,
                    self.reg_ignore,
                    self.csr_sections,
                    total = self.ref_total_instructions,
                    current = self.instructions_togen,
                    data_hazards = self.data_hazards)

            # Put the instruction
            if self.curr_inst_nolabel:
                self.q.put(('instruction_nolabel', next_inst_with_args))
                self.curr_inst_nolabel = False
            else:
                self.q.put(('instruction', next_inst_with_args))

            self.instructions_togen -= 1

        # Decrement total number of instructions
        self.total_instructions -= 1
        return

    def compute_instruction_distribution(self, args):
        """ Function to compute fraction of instructions per ISA extension """

        # Dictionary to store the values of each instruction distribution
        cd = {}

        # Count relative values of instruction set
        # config format - rel_<inst_set>_instructions
        for item in args:
            isa_index = item[0].split('_')[1]
            if float(item[1]) > 0.0:
                cd[isa_index] = float(item[1])

        # Normalize them
        total_sum = sum(list([float(k) for k in cd.values()]))
        for k in cd:
            instr_count = int(cd[k]*self.total_instructions/total_sum)
            cd[k] = int(cd[k]*self.total_instructions/total_sum)

        self.inst_dist = cd
        self.inst_dist_nobranch = {k:cd[k] for k in self.inst_dist if
                (k != 'rv32i.ctrl' and k != 'rvc.ctrl' and k != 'rv32c.ctrl')}
        if self.branch_ext_selected(self.inst_dist):
            num_branch_insts = 0
            if 'rv32i.ctrl' in self.inst_dist:
                num_branch_insts += self.inst_dist['rv32i.ctrl']
            if 'rvc.ctrl' in self.inst_dist:
                num_branch_insts += self.inst_dist['rvc.ctrl']
            if 'rv32c.ctrl' in self.inst_dist:
                num_branch_insts += self.inst_dist['rv32c.ctrl']
            try:
                self.min_insts_threshold = int(sum(self.inst_dist_nobranch.values())/num_branch_insts) - 1
                logger.info("Branch threshold: {}".format(self.min_insts_threshold))
            except:
                logger.error("Number of Instructions too low for branch. Increase Total number of instructions")
                exit(0)
        else:
            self.min_insts_threshold = 0

        self.total_instructions = sum(self.inst_dist.values())
        self.instructions_togen = self.total_instructions
        self.inst_dist_backup = self.inst_dist.copy()
        self.inst_dist_nobranch_backup = self.inst_dist_nobranch.copy()


    def init_regfile(self, not_used_reg_string):
        """ Initialize the register file | reg : (read, write) """
        # not_used_regs for all other instructions
        not_used_regs = [(x[0], int(x[1:])) for x in not_used_reg_string.strip("'").split(',')]
        if self.self_checking:
            not_used_regs.append(('x',5))
        
        # dont_use_regs used for selecting register for branch use
        dont_use_regs = [5,6,10,11,12,13,30] # 6 is for data section load, 10 is branch target

        self.rec_use_reg1 = random.randint(0,5)
        self.rec_use_reg2 = random.randint(0,5)
        while self.rec_use_reg2 == self.rec_use_reg1:
            self.rec_use_reg2 = random.randint(0,5)

        dont_use_regs.append(self.rec_use_reg1)
        dont_use_regs.append(self.rec_use_reg2)

        for i in self.no_use_regs:
            dont_use_regs.append(i)
        

        if self.branch_ext_selected(self.inst_dist):
            while(True):
                branch_use_reg = random.randint(16, 31)
                if ('x',branch_use_reg) not in not_used_regs and branch_use_reg not in dont_use_regs:
                    self.branch_use_reg = branch_use_reg
                    not_used_regs.append(('x', branch_use_reg))
                    break

        for i in range(32):
            reg = ('x', i)
            if reg not in not_used_regs:
                self.regfile[('x', i)] = (0,0)
        
        for i in range(32):
            reg = ('f', i)
            if reg not in not_used_regs:
                self.regfile[('f', i)] = (0,0)

    def add_prelude(self, args):
        """Add the prelude instructions to the queue to be written"""

        # Pre-program macro
        self.q.put(('section', 'main'))
        self.q.put(('instruction_nolabel', ('pre_program_macro', )))
        if args.getboolean('switch-priv-modes', 'switch_modes') and self.self_checking != True:
          self.q.put(('instruction_nolabel', ('la', 't0', 'switch_mode_handler')))
          self.q.put(('instruction_nolabel', ('csrw', 'mtvec', 't0')))
        elif args.getboolean('general', 'custom_trap_handler'):
            if self.delegation_boolean:
                mode = args.get('priv-mode','mode')
                if mode=="m":
                  self.q.put(('instruction_nolabel', ('la', 't0', 'custom_trap_handler')))
                  self.q.put(('instruction_nolabel', ('csrw', 'mtvec', 't0')))
                if mode=="s":
                  self.q.put(('instruction_nolabel', ('la', 't0', 'custom_trap_handler')))
                  self.q.put(('instruction_nolabel', ('csrw', 'mtvec', 't0')))
                  self.q.put(('instruction_nolabel', ('la', 't0', 'custom_trap_handler_s')))
                  self.q.put(('instruction_nolabel', ('csrw', 'stvec', 't0')))
                  self.q.put(('instruction_nolabel', ('li', 't0', '{}'.format(self.del_input))))
                  self.q.put(('instruction_nolabel', ('csrw', 'medeleg', 't0')))
                if mode=="u":
                  self.q.put(('instruction_nolabel', ('la', 't0', 'custom_trap_handler')))
                  self.q.put(('instruction_nolabel', ('csrw', 'mtvec', 't0')))
            else:
                self.q.put(('instruction_nolabel', ('la', 't0', 'custom_trap_handler')))
                self.q.put(('instruction_nolabel', ('csrw', 'mtvec', 't0')))

        self.q.put(('instruction_nolabel', ('test_entry_macro', )))
        self.q.put(('instruction_nolabel', ['123:']))
        self.total_instructions += 2

        # User trap handler
        

    def add_recursion_sections(self, depth):
        """Add user-defined templates"""
        recursion_template_enabled = self.recursion_enabled
        logger.info("Recursion Enabled? {}".format(recursion_template_enabled))

        if recursion_template_enabled:
            recurse_sections = aapg.asm_templates.recurse_sections(depth,self.rec_use_reg1,self.rec_use_reg2)
            for section in recurse_sections:
                self.q.put(('section', '.' + section))
                for instruction in recurse_sections[section]:
                    self.q.put(('pseudo', instruction))

        return

    def add_recursion_call(self):
        """ Add a recursion call """
        recursion_depth = self.args.getint('recursion-options', 'recursion-depth')
        self.q.put(('instruction', ('li', 'a0', str(recursion_depth))))
        self.q.put(('instruction', ('call', '.recurse')))
        self.total_instructions += 2
        logger.debug("Added recursion call")
        return

    def add_memory_instruction(self, random_gen = True):
        """ Add a memory instruction
            With a certain percentage, move the stack pointer within the accepted
            access sections and then check the instruction type and create the
            offset
        """
        if len(self.access_sections) == 0:
            # No access section specified
            return

        if random.random() < 0.2 or random_gen == False:
            # Select a random access section
            access_section = random.choice(self.access_sections)
            sp_address = random.randint(access_section[0] + 2048, access_section[1] - 2048)

            # Align to 64 bits
            sp_address = int(sp_address/8)*8
            self.q.put(('instruction_nolabel', ['la', 'sp', access_section[3]]))
            self.q.put(('instruction_nolabel', ['li', 't1', str(sp_address - access_section[0])]))
            self.q.put(('instruction_nolabel', ['add', 'sp', 'sp', 't1']))
            self.total_instructions += 3
            self.current_access_section = access_section

    def branch_ext_selected(self, inst_dist):
        """ Check if a branch extension is present in supplied inst dist """
        if 'rv32i.ctrl' in inst_dist or 'rvc.ctrl' in inst_dist or 'rv32c.ctrl' in inst_dist:
                return True
        else:
                return False

    def is_branch_ext(self, ext):
        if ext == 'rv32i.ctrl' or ext == 'rvc.ctrl' or ext == 'rv32c.ctrl':
            return True
        else:
            return False

    def branch_exts(self, inst_dist):
        """ Return the exts selected """
        ret = []
        for k in self.inst_dist:
            if k == 'rv32i.ctrl' or k == 'rv32c.ctrl' or k == 'rvc.ctrl':
                ret.append(k)

        return ret

class DataGenerator(object):
    """ Object to generate the data section """

    def __init__(self, size):
        """ Initialize the data generator """
        logger.debug("Data generator instantiated of size %s bytes", size)

        # Upper bound for random value in abs
        self.bounds = {
            '.dword': 1<<64 - 1,
            '.word' : 1<<32 - 1,
            '.half' : 1<<16 - 1,
            '.byte' : 1<<8 - 1
        }

        # Local variables
        if size % 64 == 0:
            # Multiple of 64 bits / dword
            size_prefix = '.dword'
            num_lines = size / 8
        elif size % 32 == 0:
            # Multiple of 32 bits / word
            size_prefix = '.word'
            num_lines = size / 4
        elif size % 16 == 0:
            # Multiple of 16 bits / half
            size_prefix = '.half'
            num_lines = size / 2
        else:
            size_prefix = '.byte'
            num_lines = size

        self.size_prefix = size_prefix
        self.num_lines = num_lines
        logger.debug("Num {}: {}".format(size_prefix, num_lines))

    def __iter__(self):
        return self

    def next(self):
        """Python 2 compat"""
        return self.__next__()

    def __next__(self):
        logger.debug("Lines left: {}".format(self.num_lines))
        if self.num_lines == 0:
            raise StopIteration("Data Section Generated")
        else:
            self.num_lines -= 1
            value = random.randint(0, self.bounds[self.size_prefix])
            address = (self.size_prefix, "{0:#0{1}x}".format(value, 18))
            return address

class ThrashGenerator(object):
    """ Object to generate the data section """

    def __init__(self, cache_type, args):
        """ Initialize the data generator """
        logger.debug("Thrash generator instantiated")

        # Read the args
        self.num_bytes_per_block = int(args.getint(cache_type, 'num_bytes_per_block'))
        self.num_blocks = int(args.getint(cache_type, 'num_blocks'))
        self.num_cycles = int(args.getint(cache_type, 'num_cycles'))
        
        # Log configured args
        logger.info("Cache thrasher type: {}".format(cache_type))
        logger.info("Number of bytes per block: {}".format(self.num_bytes_per_block))
        logger.info("Number of blocks: {}".format(self.num_blocks))
        logger.info("Number of cycles: {}".format(self.num_cycles))

        # Local variables
        self.return_address_moved = False
        self.finished = False

        # Local counters
        self.block_index = 0
        self.byte_index = 0
        self.miss = True

    def __iter__(self):
        return self

    def next(self):
        """Python 2 compat"""
        return self.__next__()

    def __next__(self):
        if self.finished:
            raise StopIteration("Thrashing section generated")

        if not self.return_address_moved:
            self.return_address_moved = True
            # Move the return address from x1 into x31
            insts = [
                ('addi', 'x31', 'x1', '0'),
            ]

            insts_left = int((self.num_bytes_per_block - 8) / 4)
            if insts_left > 0:
                insts.extend([('addi', 'x0', 'x0', '0')] * insts_left)
            return ('instructions', insts)

        # Creating the i-cache instruction
        miss = self.miss
        if self.block_index < (2 * self.num_cycles * self.num_blocks):
            if self.byte_index == 0:
                self.byte_index = 4
                if miss == True:
                    return (
                        'label',
                        'it' + '{0:09x}'.format(self.block_index),
                        ('j', 'it' + '{0:09x}'.format(self.block_index + 1)))
                elif miss == False:
                    return (
                        'label',
                        'it' + '{0:09x}'.format(self.block_index),
                        ('j', 'it' + '{0:09x} + 4'.format(self.block_index -1)))

            elif self.byte_index > 0 and self.byte_index < self.num_bytes_per_block - 4:
                self.byte_index += 4
                nop = ('instruction',('nop',))
                return nop
            elif self.byte_index == self.num_bytes_per_block - 4:
                self.byte_index += 4
                if miss == True:
                    return (
                        'instruction',
                        ('j', 'it' + '{0:09x} + 4'.format(self.block_index +1))
                    )
                elif miss == False:
                    nop = ('instruction',('nop',))
                    return nop
            elif self.byte_index == self.num_bytes_per_block:
                self.byte_index = 0
                self.block_index += 1
                self.miss = not self.miss
        elif self.block_index == 2 * self.num_cycles * self.num_blocks:
            self.finished = True
            return (
                'label_instructions',
                'it' + '{0:09x}'.format(self.block_index),
                [
                    ('jr', 'x31'),
                ])



import numpy as np

templates_asm = ['''
#define PRV_S 1
#define PRV_U 0
#define MSTATUS_MPP 0x00001800
#if __riscv_xlen == 64
  #define LREG ld
  #define SREG sd
  #define LREGU lwu
  #define REGBYTES 8
  #define FMV fmv.d.x
#else
  #define LREG lw
  #define LREGU lw
  #define SREG sw
  #define REGBYTES 4
  #define FMV fmv.w.x
#endif


# From here 
.align 2
.globl custom_trap_handler_s
custom_trap_handler_s:
  addi sp, sp, -40*REGBYTES

  SREG x1, 1*REGBYTES(sp)
  SREG x3, 3*REGBYTES(sp)
  SREG x4, 4*REGBYTES(sp)
  SREG x5, 5*REGBYTES(sp)
  SREG x6, 6*REGBYTES(sp)
  SREG x7, 7*REGBYTES(sp)
  SREG x8, 8*REGBYTES(sp)
  SREG x9, 9*REGBYTES(sp)
  SREG x10, 10*REGBYTES(sp)
  SREG x11, 11*REGBYTES(sp)
  SREG x12, 12*REGBYTES(sp)
  SREG x13, 13*REGBYTES(sp)
  SREG x14, 14*REGBYTES(sp)
  SREG x15, 15*REGBYTES(sp)
  SREG x16, 16*REGBYTES(sp)
  SREG x17, 17*REGBYTES(sp)
  SREG x18, 18*REGBYTES(sp)
  SREG x19, 19*REGBYTES(sp)
  SREG x20, 20*REGBYTES(sp)
  SREG x21, 21*REGBYTES(sp)
  SREG x22, 22*REGBYTES(sp)
  SREG x23, 23*REGBYTES(sp)
  SREG x24, 24*REGBYTES(sp)
  SREG x25, 25*REGBYTES(sp)
  SREG x26, 26*REGBYTES(sp)
  SREG x27, 27*REGBYTES(sp)
  SREG x28, 28*REGBYTES(sp)
  SREG x29, 29*REGBYTES(sp)
  SREG x30, 30*REGBYTES(sp)
  SREG x31, 31*REGBYTES(sp)
  csrr a0, sstatus
  csrr a0, sepc
  csrr a0, stvec
  csrr a0, scause 
  li a2, 0
  beq a0, a2, 1f
  li a2, 1
  beq a0, a2, 1f
  csrr a1, sepc   
  li a2, 2
  beq a0, a2, inst32_s
1:
  lh a2, (a1)
  # check the lower 2 bits to see if the instruction is 32-bit or 16-bit.
  andi a2, a2, 0x3;
  li t0, 0x3
  bne a2,t0,inst16_s
inst32_s:                           # is 32-bit instruction then increment by 4
  addi a1,a1,0x4
  beqz x0,1f
inst16_s:
  addi a1,a1,0x2                  # is 16-bit instruction then increment by 2
1: 
  csrw sepc, a1                   # point mepc to the next instruction.


  LREG x1, 1*REGBYTES(sp)
  LREG x3, 3*REGBYTES(sp)
  LREG x4, 4*REGBYTES(sp)
  LREG x5, 5*REGBYTES(sp)
  LREG x6, 6*REGBYTES(sp)
  LREG x7, 7*REGBYTES(sp)
  LREG x8, 8*REGBYTES(sp)
  LREG x9, 9*REGBYTES(sp)
  LREG x10, 10*REGBYTES(sp)
  LREG x11, 11*REGBYTES(sp)
  LREG x12, 12*REGBYTES(sp)
  LREG x13, 13*REGBYTES(sp)
  LREG x14, 14*REGBYTES(sp)
  LREG x15, 15*REGBYTES(sp)
  LREG x16, 16*REGBYTES(sp)
  LREG x17, 17*REGBYTES(sp)
  LREG x18, 18*REGBYTES(sp)
  LREG x19, 19*REGBYTES(sp)
  LREG x20, 20*REGBYTES(sp)
  LREG x21, 21*REGBYTES(sp)
  LREG x22, 22*REGBYTES(sp)
  LREG x23, 23*REGBYTES(sp)
  LREG x24, 24*REGBYTES(sp)
  LREG x25, 25*REGBYTES(sp)
  LREG x26, 26*REGBYTES(sp)
  LREG x27, 27*REGBYTES(sp)
  LREG x28, 28*REGBYTES(sp)
  LREG x29, 29*REGBYTES(sp)
  LREG x30, 30*REGBYTES(sp)
  LREG x31, 31*REGBYTES(sp)
  addi sp, sp, 40*REGBYTES
  sret

.align 2
.globl custom_trap_handler
custom_trap_handler:
  addi sp, sp, -40*REGBYTES

  SREG x1, 1*REGBYTES(sp)
  SREG x3, 3*REGBYTES(sp)
  SREG x4, 4*REGBYTES(sp)
  SREG x5, 5*REGBYTES(sp)
  SREG x6, 6*REGBYTES(sp)
  SREG x7, 7*REGBYTES(sp)
  SREG x8, 8*REGBYTES(sp)
  SREG x9, 9*REGBYTES(sp)
  SREG x10, 10*REGBYTES(sp)
  SREG x11, 11*REGBYTES(sp)
  SREG x12, 12*REGBYTES(sp)
  SREG x13, 13*REGBYTES(sp)
  SREG x14, 14*REGBYTES(sp)
  SREG x15, 15*REGBYTES(sp)
  SREG x16, 16*REGBYTES(sp)
  SREG x17, 17*REGBYTES(sp)
  SREG x18, 18*REGBYTES(sp)
  SREG x19, 19*REGBYTES(sp)
  SREG x20, 20*REGBYTES(sp)
  SREG x21, 21*REGBYTES(sp)
  SREG x22, 22*REGBYTES(sp)
  SREG x23, 23*REGBYTES(sp)
  SREG x24, 24*REGBYTES(sp)
  SREG x25, 25*REGBYTES(sp)
  SREG x26, 26*REGBYTES(sp)
  SREG x27, 27*REGBYTES(sp)
  SREG x28, 28*REGBYTES(sp)
  SREG x29, 29*REGBYTES(sp)
  SREG x30, 30*REGBYTES(sp)
  SREG x31, 31*REGBYTES(sp)
  csrr a0, mstatus
  csrr a0, mepc
  csrr a0, mtvec
  csrr a0, mcause 
  li a2, 0
  beq a0, a2, 1f
  li a2, 1
  beq a0, a2, 1f
  csrr a1, mepc   
  li a2, 2
  beq a0, a2, inst32
1:
  lh a2, (a1)
  # check the lower 2 bits to see if the instruction is 32-bit or 16-bit.
  andi a2, a2, 0x3;
  li t0, 0x3
  bne a2,t0,inst16
inst32:                           # is 32-bit instruction then increment by 4
  addi a1,a1,0x4
  beqz x0,1f
inst16:
  addi a1,a1,0x2                  # is 16-bit instruction then increment by 2
1: 
  csrw mepc, a1                   # point mepc to the next instruction.


  LREG x1, 1*REGBYTES(sp)
  LREG x3, 3*REGBYTES(sp)
  LREG x4, 4*REGBYTES(sp)
  LREG x5, 5*REGBYTES(sp)
  LREG x6, 6*REGBYTES(sp)
  LREG x7, 7*REGBYTES(sp)
  LREG x8, 8*REGBYTES(sp)
  LREG x9, 9*REGBYTES(sp)
  LREG x10, 10*REGBYTES(sp)
  LREG x11, 11*REGBYTES(sp)
  LREG x12, 12*REGBYTES(sp)
  LREG x13, 13*REGBYTES(sp)
  LREG x14, 14*REGBYTES(sp)
  LREG x15, 15*REGBYTES(sp)
  LREG x16, 16*REGBYTES(sp)
  LREG x17, 17*REGBYTES(sp)
  LREG x18, 18*REGBYTES(sp)
  LREG x19, 19*REGBYTES(sp)
  LREG x20, 20*REGBYTES(sp)
  LREG x21, 21*REGBYTES(sp)
  LREG x22, 22*REGBYTES(sp)
  LREG x23, 23*REGBYTES(sp)
  LREG x24, 24*REGBYTES(sp)
  LREG x25, 25*REGBYTES(sp)
  LREG x26, 26*REGBYTES(sp)
  LREG x27, 27*REGBYTES(sp)
  LREG x28, 28*REGBYTES(sp)
  LREG x29, 29*REGBYTES(sp)
  LREG x30, 30*REGBYTES(sp)
  LREG x31, 31*REGBYTES(sp)
  addi sp, sp, 40*REGBYTES
  mret


# To here
# Instructions to be inserted before and after the program
# Possible setups inside this macro can be
#   1. Privilege mode change
#   2. PMP Configuration - csrw pmpcfg0 xxx, csrw pmpcfg1 xxx

############################
### Exception generation ###
############################

# Instruction address misaligned
.macro ecause00
.endm

# Instruction access fault
.macro ecause01
li x15, -10
jr x15
.endm

# Illegal Instruction
.macro ecause02
.word 0x01239239812981
.endm

# Breakpoint
.macro ecause03
ebreak
.endm

# Load address misaligned
.macro ecause04
LREG x0, (REGBYTES-2)(sp)
.endm

# Load access fault
.macro ecause05
.endm

# Store/AMO address misaligned
.macro ecause06
SREG x0, (REGBYTES-2)(sp)
.endm

# Store/AMO access fault
.macro ecause07
.endm

# Env call from U-mode
.macro ecause08
.endm

# Env call from S-mode
.macro ecause09
.endm

# Reserved
.macro ecause10
.endm

# Env call from M-mode
.macro ecause11
ecall
.endm

# Instruction page fault
.macro ecause12
.endm

# Load page fault
.macro ecause13
.endm

# Reserved
.macro ecause14
.endm

# Store/AMO page fault
.macro ecause15
.endm

# Env call 
.macro switchmodes
ecall
.endm
''','''
#define PRV_S 1
#define PRV_U 0
#define MSTATUS_MPP 0x00001800
#define MSTATUS_FS          0x00006000
#define RVTEST_FP_ENABLE                                                \
  li a0, MSTATUS_FS & (MSTATUS_FS >> 1);                                \
  csrs mstatus, a0;                                                     \
  csrwi fcsr, 0
#if __riscv_xlen == 64
  #define LREG ld
  #define SREG sd
  #define LREGU lwu
  #define REGBYTES 8
  #define FMV fmv.d.x
#else
  #define LREG lw
  #define LREGU lw
  #define SREG sw
  #define REGBYTES 4
  #define FMV fmv.w.x
#endif


# From here 
.align 2
.globl switch_mode_handler
switch_mode_handler:
  addi sp, sp, -40*REGBYTES
  SREG x1, 1*REGBYTES(sp)
  SREG x3, 3*REGBYTES(sp)
  SREG x4, 4*REGBYTES(sp)
  SREG x5, 5*REGBYTES(sp)
  SREG x6, 6*REGBYTES(sp)
  SREG x7, 7*REGBYTES(sp)
  SREG x8, 8*REGBYTES(sp)
  SREG x9, 9*REGBYTES(sp)
  SREG x10, 10*REGBYTES(sp)
  SREG x11, 11*REGBYTES(sp)
  SREG x12, 12*REGBYTES(sp)
  SREG x13, 13*REGBYTES(sp)
  SREG x14, 14*REGBYTES(sp)
  SREG x15, 15*REGBYTES(sp)
  SREG x16, 16*REGBYTES(sp)
  SREG x17, 17*REGBYTES(sp)
  SREG x18, 18*REGBYTES(sp)
  SREG x19, 19*REGBYTES(sp)
  SREG x20, 20*REGBYTES(sp)
  SREG x21, 21*REGBYTES(sp)
  SREG x22, 22*REGBYTES(sp)
  SREG x23, 23*REGBYTES(sp)
  SREG x24, 24*REGBYTES(sp)
  SREG x25, 25*REGBYTES(sp)
  SREG x26, 26*REGBYTES(sp)
  SREG x27, 27*REGBYTES(sp)
  SREG x28, 28*REGBYTES(sp)
  SREG x29, 29*REGBYTES(sp)
  SREG x30, 30*REGBYTES(sp)
  SREG x31, 31*REGBYTES(sp)
  csrr a0, mstatus
  csrr a0, mepc
  csrr a0, mtvec 
  csrr a0, mcause
 
  li a2, 0
  beq a0, a2, 1f
  li a2, 1
  beq a0, a2, 1f
  csrr a1, mepc   
  li a2, 2
  beq a0, a2, inst32_2
1:
  lh a2, (a1)
  # check the lower 2 bits to see if the instruction is 32-bit or 16-bit.
  andi a2, a2, 0x3;
  li t0, 0x3
  bne a2,t0,inst16_2
inst32_2:                           # is 32-bit instruction then increment by 4
  addi a1,a1,0x4
  beqz x0,1f
inst16_2:
  addi a1,a1,0x2                  # is 16-bit instruction then increment by 2

1: 
  csrw mepc, a1                   # point mepc to the next instruction.
  csrr a3, mcause
  li t1, 8
  beq a3,t1, sw1
  li t1, 9
  beq a3,t1, sw1
  li t1, 11
  beq a3,t1, sw1
  beq x0, x0, 1f
  mret

sw1:
  csrr a3,mepc
  andi a3,a3,12
  
  li t1, 0
  beq a3,t1, eq1 # Jump to supervisor

  li t1, 4
  beq a3,t1, eq2 # Continue in same mode

  li t1, 8
  beq a3,t1, eq3 # Jump to machine

  li t1, 12
  beq a3,t1, eq4 # Jump to user  
  beq x0, x0, 1f


eq1:
  li t0, MSTATUS_MPP
  csrc mstatus, t0
  li t5, (MSTATUS_MPP & -MSTATUS_MPP) * PRV_S
  csrs mstatus, t5
  RVTEST_FP_ENABLE;
  beq x0, x0, 1f
eq2:
  RVTEST_FP_ENABLE;
  beq x0, x0, 1f
  
eq3:
  li t0, 0x200001800
  csrw 0x300, t0 # MSTATUS
  li t0, 0x0
  csrw 0x304, t0 # MIE
  RVTEST_FP_ENABLE;
  beq x0, x0, 1f

eq4:
  li t0, MSTATUS_MPP
  csrc mstatus, t0
  li t5, (MSTATUS_MPP & -MSTATUS_MPP) * PRV_U
  csrs mstatus, t5
  RVTEST_FP_ENABLE;
  beq x0, x0, 1f

1:
  LREG x1, 1*REGBYTES(sp)
  LREG x3, 3*REGBYTES(sp)
  LREG x4, 4*REGBYTES(sp)
  LREG x5, 5*REGBYTES(sp)
  LREG x6, 6*REGBYTES(sp)
  LREG x7, 7*REGBYTES(sp)
  LREG x8, 8*REGBYTES(sp)
  LREG x9, 9*REGBYTES(sp)
  LREG x10, 10*REGBYTES(sp)
  LREG x11, 11*REGBYTES(sp)
  LREG x12, 12*REGBYTES(sp)
  LREG x13, 13*REGBYTES(sp)
  LREG x14, 14*REGBYTES(sp)
  LREG x15, 15*REGBYTES(sp)
  LREG x16, 16*REGBYTES(sp)
  LREG x17, 17*REGBYTES(sp)
  LREG x18, 18*REGBYTES(sp)
  LREG x19, 19*REGBYTES(sp)
  LREG x20, 20*REGBYTES(sp)
  LREG x21, 21*REGBYTES(sp)
  LREG x22, 22*REGBYTES(sp)
  LREG x23, 23*REGBYTES(sp)
  LREG x24, 24*REGBYTES(sp)
  LREG x25, 25*REGBYTES(sp)
  LREG x26, 26*REGBYTES(sp)
  LREG x27, 27*REGBYTES(sp)
  LREG x28, 28*REGBYTES(sp)
  LREG x29, 29*REGBYTES(sp)
  LREG x30, 30*REGBYTES(sp)
  LREG x31, 31*REGBYTES(sp)
  addi sp, sp, 40*REGBYTES
  mret


# To here
# Instructions to be inserted before and after the program
# Possible setups inside this macro can be
#   1. Privilege mode change
#   2. PMP Configuration - csrw pmpcfg0 xxx, csrw pmpcfg1 xxx

############################
### Exception generation ###
############################

# Instruction address misaligned
.macro ecause00
.endm

# Instruction access fault
.macro ecause01
li x15, -10
jr x15
.endm

# Illegal Instruction
.macro ecause02
.word 0x01239239812981
.endm

# Breakpoint
.macro ecause03
ebreak
.endm

# Load address misaligned
.macro ecause04
LREG x0, (REGBYTES-2)(sp)
.endm

# Load access fault
.macro ecause05
.endm

# Store/AMO address misaligned
.macro ecause06
SREG x0, (REGBYTES-2)(sp)
.endm

# Store/AMO access fault
.macro ecause07
.endm

# Env call from U-mode
.macro ecause08
.endm

# Env call from S-mode
.macro ecause09
.endm

# Reserved
.macro ecause10
.endm

# Env call from M-mode
.macro ecause11
ecall
.endm

# Instruction page fault
.macro ecause12
.endm

# Load page fault
.macro ecause13
.endm

# Reserved
.macro ecause14
.endm

# Store/AMO page fault
.macro ecause15
.endm

# Env call 
.macro switchmodes
ecall
.endm
'''
]



"""
    Automatic Assembly Program Generator (AAPG)

    Module of AAPG that deals with
        * Checking if signature file is created
        * Parsing signature file
        * Appending parseg signature file to test
        * Appending test function to template.S
"""
import logging
import argparse
import os
import io
import sys
import subprocess
import click
import yaml
import fileinput
from glob import glob

def setup_logging(log_level):
    """Setup logging

        Verbosity decided on user input

        Args:
            log_level: (str) User defined log level

        Returns:
            None
    """
    numeric_level = getattr(logging, log_level.upper(), None)

    if not isinstance(numeric_level, int):
        print("\033[91mInvalid log level passed. Please select from debug | info | warning | error\033[0m")
        sys.exit(1)

    logging.basicConfig(level = numeric_level)


def format_sig(sig_file):
    """Format Signature

        Function to parse the generated signature files and return as list
    """
    f = open(sig_file,'r')
    to_write = []
    while(True):
        line = f.readline()
        if not line:
            break
        write_line2,write_line1 = line[0:16],line[16:32]
        write_line1 = '.dword 0x'+write_line1
        write_line2 = '.dword 0x'+write_line2
        to_write.append(write_line1)
        to_write.append(write_line2)
    f.close()
    return to_write

def add_function_call(test_file,x):
    if x == 2:
        for line in fileinput.input(test_file, inplace=True):
            if "post_program_macro" in line:
                if "#" not in line:
                    print("test_pass_macro\n", end='')
                    print(line, end='')
            elif "write_chsum" in line:
                line = line.replace("write_chsum","read_chsum")
                print(line, end='')
            else:
                print(line, end='')

    if x == 1:
        for line in fileinput.input(test_file, inplace=True):
            if "post_program_macro" in line:
                if "#" not in line:
                    print("call write_chsum\n", end='')
                    print(line, end='')
            else:
                print(line, end='')

def change_func_def(template_file):
    for line in fileinput.input(template_file, inplace=True):
        if "la                  sp, end_signature" in line:
            if "#" not in line:
                print("        la                  sp, end_ref_signature\n", end='')
        else:
            print(line, end='')


def add_self_check(setup_dir,output_dir,config_file):
    """Add Self Check

        Function to execute the makefile, extract signature and append to test

        Args:

    """
    setup_logging('info')
    logger = logging.getLogger()

    logger.info('Executing Makefile to generate signature dump \n')


    assembly = [y for x in os.walk(os.path.join(setup_dir,'asm')) for y in glob(os.path.join(x[0], '*.S'))]

    test_file = None
    template_file = None
    for file in assembly:
        if "template.S" not in file:
            test_file = file
        elif "template.S" in file:
            template_file = file

    # Calculate the size of the signature section
    # Still in development. Code to check if Data section is same
    conf = yaml.safe_load(open(os.path.join(setup_dir,config_file)))
    signature_start = None
    signature_end = None
    try:
        signature_section = conf['access-sections']['begin_signature']
        signature = signature_section.split(',')
        signature_start = int(signature[0],16)
        signature_end = int(signature[1],16)
    except:
        logger.error("Can not find `begin_signature` section in config file.")
        sys.exit(0)

    length = int((signature_end - signature_start)/8)
    logger.info("Size of signature section is {}".format(str(length)))
    length = length - 1

    func_to_add = """
.globl check_sign_func        
check_sign_func:
li t0, {}*REGBYTES
loop_label:
  la      sp, begin_signature
  add     sp, sp, t0
  LREG    t1, 0*REGBYTES(sp)

  la      sp, ref_signature
  add     sp, sp, t0
  LREG    t2, 0*REGBYTES(sp)

  bne     t1, t2, fail
  addi    t0, t0, -1*REGBYTES
  bgtz    t0, loop_label
pass:
  la      sp, begin_signature
  addi    sp, sp, 0 
  addi    t1, x0, 1
  SREG    t1, 0*REGBYTES(sp)
  ret
fail:
  la      sp, begin_signature
  addi    sp, sp, 0 
  addi    t1, x0, 0
  SREG    t1, 0*REGBYTES(sp)
  ret
    """.format(str(length))

    func_to_add2 = """
.option push;
.option norvc;
.globl write_chsum        
write_chsum:  
        csrw mscratch,   sp                 # store the sp in mscratch
        la   sp,  register_swap
        SREG x3, 1*REGBYTES(sp)
        SREG x4, 2*REGBYTES(sp)
        SREG x5, 3*REGBYTES(sp)
        SREG x6, 4*REGBYTES(sp)
        SREG x7, 5*REGBYTES(sp)
        SREG x8, 6*REGBYTES(sp)
        SREG x9, 7*REGBYTES(sp)
        SREG x10, 8*REGBYTES(sp)
        SREG x11, 9*REGBYTES(sp)
        SREG x12, 10*REGBYTES(sp)
        SREG x13, 11*REGBYTES(sp)
        SREG x14, 12*REGBYTES(sp)
        SREG x15, 13*REGBYTES(sp)
        SREG x16, 14*REGBYTES(sp)
        SREG x17, 15*REGBYTES(sp)
        SREG x18, 16*REGBYTES(sp)
        SREG x19, 17*REGBYTES(sp)
        SREG x20, 18*REGBYTES(sp)
        SREG x21, 19*REGBYTES(sp)
        SREG x22, 20*REGBYTES(sp)
        SREG x23, 21*REGBYTES(sp)
        SREG x24, 22*REGBYTES(sp)
        SREG x25, 23*REGBYTES(sp)
        SREG x26, 24*REGBYTES(sp)
        SREG x27, 25*REGBYTES(sp)
        SREG x28, 26*REGBYTES(sp)
        SREG x29, 27*REGBYTES(sp)
        SREG x30, 28*REGBYTES(sp)
        SREG x31, 29*REGBYTES(sp)
        SREG x1, 30*REGBYTES(sp)
        SREG x0, 31*REGBYTES(sp)
        csrr x30, mscratch                  # copy orig-sp in x30
        SREG x30, 32*REGBYTES(sp)           # store orig-sp
        li                  t6,  0
        li                  t4,  0
        li                  t3,  33
wnext_sum:
        addi                t4, t4, 1
        beq                 t4, t3, wfinchsum
        LREG                t5,  1*REGBYTES(sp)
        addi                sp, sp, 1*REGBYTES
        add                 t6, t6, t5
        bgt                 t6, t5, wnext_sum
        addi                t6, t6, 1
        beq                 x0, x0, wnext_sum 
wfinchsum:
        addi                t0, t0, -1*REGBYTES
        la                  sp, end_signature
        add                 sp, sp, t0
        SREG                t6, 0*REGBYTES(sp)
        la                  sp,  register_swap
        LREG x3, 1*REGBYTES(sp)
        LREG x4, 2*REGBYTES(sp)
        LREG x6, 4*REGBYTES(sp)
        LREG x7, 5*REGBYTES(sp)
        LREG x8, 6*REGBYTES(sp)
        LREG x9, 7*REGBYTES(sp)
        LREG x10, 8*REGBYTES(sp)
        LREG x11, 9*REGBYTES(sp)
        LREG x12, 10*REGBYTES(sp)
        LREG x13, 11*REGBYTES(sp)
        LREG x14, 12*REGBYTES(sp)
        LREG x15, 13*REGBYTES(sp)
        LREG x16, 14*REGBYTES(sp)
        LREG x17, 15*REGBYTES(sp)
        LREG x18, 16*REGBYTES(sp)
        LREG x19, 17*REGBYTES(sp)
        LREG x20, 18*REGBYTES(sp)
        LREG x21, 19*REGBYTES(sp)
        LREG x22, 20*REGBYTES(sp)
        LREG x23, 21*REGBYTES(sp)
        LREG x24, 22*REGBYTES(sp)
        LREG x25, 23*REGBYTES(sp)
        LREG x26, 24*REGBYTES(sp)
        LREG x27, 25*REGBYTES(sp)
        LREG x28, 26*REGBYTES(sp)
        LREG x29, 27*REGBYTES(sp)
        LREG x30, 28*REGBYTES(sp)
        LREG x31, 29*REGBYTES(sp)
        LREG x1, 30*REGBYTES(sp)
        csrr    sp, mscratch
        ret
.option pop;
    """

    func_to_add3 = """
.option push;
.option norvc;
.globl read_chsum        
read_chsum:       
        csrw                mscratch,   sp
        la                  sp,  register_swap
        SREG x3, 1*REGBYTES(sp)
        SREG x4, 2*REGBYTES(sp)
        SREG x5, 3*REGBYTES(sp)
        SREG x6, 4*REGBYTES(sp)
        SREG x7, 5*REGBYTES(sp)
        SREG x8, 6*REGBYTES(sp)
        SREG x9, 7*REGBYTES(sp)
        SREG x10, 8*REGBYTES(sp)
        SREG x11, 9*REGBYTES(sp)
        SREG x12, 10*REGBYTES(sp)
        SREG x13, 11*REGBYTES(sp)
        SREG x14, 12*REGBYTES(sp)
        SREG x15, 13*REGBYTES(sp)
        SREG x16, 14*REGBYTES(sp)
        SREG x17, 15*REGBYTES(sp)
        SREG x18, 16*REGBYTES(sp)
        SREG x19, 17*REGBYTES(sp)
        SREG x20, 18*REGBYTES(sp)
        SREG x21, 19*REGBYTES(sp)
        SREG x22, 20*REGBYTES(sp)
        SREG x23, 21*REGBYTES(sp)
        SREG x24, 22*REGBYTES(sp)
        SREG x25, 23*REGBYTES(sp)
        SREG x26, 24*REGBYTES(sp)
        SREG x27, 25*REGBYTES(sp)
        SREG x28, 26*REGBYTES(sp)
        SREG x29, 27*REGBYTES(sp)
        SREG x30, 28*REGBYTES(sp)
        SREG x31, 29*REGBYTES(sp)
        SREG x1, 30*REGBYTES(sp)
        SREG x0, 31*REGBYTES(sp)
        csrr x30, mscratch
        SREG x30, 32*REGBYTES(sp)
        li                  t6,  0
        li                  t4,  0
        li                  t3,  33
next_sum:
        addi                t4, t4, 1
        beq                 t4, t3, finchsum
        LREG                t5,  1*REGBYTES(sp)
        addi                sp, sp, 1*REGBYTES
        add                 t6, t6, t5
        bgt                 t6, t5, next_sum
        addi                t6, t6, 1
        beq                 x0, x0, next_sum
finchsum:
        addi                t0, t0, -1*REGBYTES
        la                  sp, end_signature
        add                 sp, sp, t0
        LREG                t2, 0*REGBYTES(sp)
        bne                 t6, t2, chfail
chpass:
        # la      sp, begin_signature
        # addi    sp, sp, 2*REGBYTES
        # li      t1, 0xfff
        # SREG    t1, 0*REGBYTES(sp)
        la                  sp,  register_swap
        LREG x3, 1*REGBYTES(sp)
        LREG x4, 2*REGBYTES(sp)
        LREG x6, 4*REGBYTES(sp)
        LREG x7, 5*REGBYTES(sp)
        LREG x8, 6*REGBYTES(sp)
        LREG x9, 7*REGBYTES(sp)
        LREG x10, 8*REGBYTES(sp)
        LREG x11, 9*REGBYTES(sp)
        LREG x12, 10*REGBYTES(sp)
        LREG x13, 11*REGBYTES(sp)
        LREG x14, 12*REGBYTES(sp)
        LREG x15, 13*REGBYTES(sp)
        LREG x16, 14*REGBYTES(sp)
        LREG x17, 15*REGBYTES(sp)
        LREG x18, 16*REGBYTES(sp)
        LREG x19, 17*REGBYTES(sp)
        LREG x20, 18*REGBYTES(sp)
        LREG x21, 19*REGBYTES(sp)
        LREG x22, 20*REGBYTES(sp)
        LREG x23, 21*REGBYTES(sp)
        LREG x24, 22*REGBYTES(sp)
        LREG x25, 23*REGBYTES(sp)
        LREG x26, 24*REGBYTES(sp)
        LREG x27, 25*REGBYTES(sp)
        LREG x28, 26*REGBYTES(sp)
        LREG x29, 27*REGBYTES(sp)
        LREG x30, 28*REGBYTES(sp)
        LREG x31, 29*REGBYTES(sp)
        LREG x1, 30*REGBYTES(sp)
        csrr sp, mscratch
        ret
chfail:
        la      sp, begin_signature
        addi    sp, sp, 2*REGBYTES
        sub     t0, x0, t0
        li      t1, 1*REGBYTES
        div     t0, t0, t1
        SREG    t0, 0*REGBYTES(sp)
        test_fail_macro
        post_program_macro
        ret
.option pop;
    """

    logger.info("Adding validation function to template file")

    check_function = func_to_add.split("\n")
    f = open(template_file,'a+')
    f.writelines(["%s\n" % item  for item in check_function])
    f.close()

    write_chsum_function = func_to_add2.split("\n")
    f = open(template_file,'a+')
    f.writelines(["%s\n" % item  for item in write_chsum_function])
    f.close()

    read_chsum_function = func_to_add3.split("\n")
    f = open(template_file,'a+')
    f.writelines(["%s\n" % item  for item in read_chsum_function])
    f.close()

    to_write1 = ['\t.data','\t.align 1','\t.globl ref_signature','ref_signature:']
    f = open(test_file,'a+')
    f.writelines(["%s\n" % item  for item in to_write1])
    f.close()
    
    # Execute MakeFile
    os.system("cd {};make".format(setup_dir))

    # Check if signature file exists
    signature = [y for x in os.walk(setup_dir) for y in glob(os.path.join(x[0], '*.dump.sign'))]
    if not signature:
        logger.error("Signature File was not created; Check if signature section is present in config/assembly")
        sys.exit(0)
    to_write = format_sig(signature[0])
    

    logger.info("Adding signature to the test file")

    f = open(test_file,'a+')
    f.writelines(["%s\n" % item  for item in to_write])
    f.close()

    to_write1 = ['\t.data','\t.align 1','\t.globl end_ref_signature','end_ref_signature:']
    f = open(test_file,'a+')
    f.writelines(["%s\n" % item  for item in to_write1])
    f.close()

    logger.info("Adding function call to test file")

    add_function_call(test_file,2)    
    change_func_def(template_file)
    os.system("cd {}; make clean".format(setup_dir))



def make_format_func(isa_string,abi_string):
	makefile = '''
XLEN ?= 64
TARGET ?= unknown-elf
RISCVPREFIX=riscv${XLEN}-${TARGET}
ASM_SRC_DIR := asm
COMMON_DIR := common
BIN_DIR := bin
OBJ_DIR := objdump
LOG_DIR := log
ISA ?= {isa_string}
ABI ?= {abi_string}

INCLUDE_DIRS := common
CRT_FILE := common/crt.S
TEMPLATE_FILE := common/templates.S
GCC_OPTS := -march=$(ISA) -mabi=$(ABI) -DPREALLOCATE=1 -mcmodel=medany -static -std=gnu99 -O2 -fno-common -fno-builtin-printf
LINKER_OPTIONS := -static -nostdlib -nostartfiles -lm -lgcc -T 

BASE_SRC_FILES := $(wildcard $(ASM_SRC_DIR)/*.S)
SRC_FILES := $(filter-out $(wildcard $(ASM_SRC_DIR)/*template.S),$(BASE_SRC_FILES))
BIN_FILES := $(patsubst $(ASM_SRC_DIR)/%.S, $(BIN_DIR)/%.riscv, $(SRC_FILES))
OBJ_FILES := $(patsubst $(ASM_SRC_DIR)/%.S, $(OBJ_DIR)/%.objdump, $(SRC_FILES))
LOG_FILES := $(patsubst $(ASM_SRC_DIR)/%.S, $(LOG_DIR)/%.log, $(SRC_FILES))
DUMP_FILES := $(patsubst $(ASM_SRC_DIR)/%.S, $(LOG_DIR)/%.dump, $(SRC_FILES))

all: build objdump run dump
\t$(info ==================== Complete Build Finished =============)

build: $(BIN_FILES)
\t$(info ==================== Build completed ====================)
\t$(info )

$(BIN_DIR)/%.riscv: $(ASM_SRC_DIR)/%.S 
\t$(info ==================== Compiling asm to binary ============)
\t${RISCVPREFIX}-gcc $(GCC_OPTS) -I $(INCLUDE_DIRS) -o $@ $< $(CRT_FILE) $(LINKER_OPTIONS) $(<D)/$*.ld


objdump: $(OBJ_FILES)
\t$(info ==================== Objdump Completed ==================)
\t$(info )

$(OBJ_DIR)/%.objdump: $(BIN_DIR)/%.riscv
\t$(info ==================== Disassembling binary ===============)
\t${RISCVPREFIX}-objdump -D $< > $@

run: $(LOG_FILES)
\t$(info ==================== Spike Run Completed ================)
\t$(info )

$(LOG_DIR)/%.log: $(BIN_DIR)/%.riscv
\t$(info ==================== Simulating binary on Spike =========)
\tspike -l --isa=$(ISA) $< 2> $@

dump: $(DUMP_FILES)
\t$(info ==================== Spike dump Completed ===============)
\t$(info )

$(LOG_DIR)/%.dump: $(BIN_DIR)/%.riscv
\t$(info ==================== Generating spike dump ==============)
\tspike --log-commits --log  $@ --isa=$(ISA) +signature=$@.sign $<



.PHONY: clean
clean:
\trm -rf bin/* log/* objdump/*
'''.format(XLEN="{XLEN}",TARGET="{TARGET}",RISCVPREFIX="{RISCVPREFIX}",isa_string=isa_string,abi_string=abi_string)
	return makefile




import pytest 
import os
import random
import numpy as np
from random import seed
from random import randint
from click.testing import CliRunner
from aapg.main import cli

seeds = []
seed(1)
for _ in range(30):
    value = randint(0, 1000)
    seeds.append(value)

@pytest.fixture
def runner():
    return CliRunner()

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_hazards_s(runner):
    """Testing aapg_iclass_rv64imafdc_hazards_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_hazards_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_hazards_s.yaml tests/work/aapg_iclass_rv64imafdc_hazards_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_hazards_s','--output_dir=tests/work/aapg_iclass_rv64imafdc_hazards_s/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_hazards_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_recurse_med(runner):
    """Testing aapg_iclass_rv64imafdc_recurse_med.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_recurse_med'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_recurse_med.yaml tests/work/aapg_iclass_rv64imafdc_recurse_med/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_recurse_med','--output_dir=tests/work/aapg_iclass_rv64imafdc_recurse_med/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_recurse_med; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_hazards_u(runner):
    """Testing aapg_cclass_rv64imafdc_hazards_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_hazards_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_hazards_u.yaml tests/work/aapg_cclass_rv64imafdc_hazards_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_hazards_u','--output_dir=tests/work/aapg_cclass_rv64imafdc_hazards_u/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_hazards_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_test_all2(runner):
    """Testing aapg_cclass_rv64imafdc_test_all2.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_test_all2'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_test_all2.yaml tests/work/aapg_cclass_rv64imafdc_test_all2/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_test_all2','--output_dir=tests/work/aapg_cclass_rv64imafdc_test_all2/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_test_all2; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_branches1(runner):
    """Testing aapg_cclass_rv64imafdc_branches1.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_branches1'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_branches1.yaml tests/work/aapg_cclass_rv64imafdc_branches1/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_branches1','--output_dir=tests/work/aapg_cclass_rv64imafdc_branches1/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_branches1; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_bringup_s(runner):
    """Testing aapg_cclass_rv64imafdc_bringup_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_bringup_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_bringup_s.yaml tests/work/aapg_cclass_rv64imafdc_bringup_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_bringup_s','--output_dir=tests/work/aapg_cclass_rv64imafdc_bringup_s/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_bringup_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_s_recursion(runner):
    """Testing aapg_cclass_rv64imafdc_s_recursion.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_s_recursion'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_s_recursion.yaml tests/work/aapg_cclass_rv64imafdc_s_recursion/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_s_recursion','--output_dir=tests/work/aapg_cclass_rv64imafdc_s_recursion/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_s_recursion; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_branches1_u(runner):
    """Testing aapg_iclass_rv64imafdc_branches1_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_branches1_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_branches1_u.yaml tests/work/aapg_iclass_rv64imafdc_branches1_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_branches1_u','--output_dir=tests/work/aapg_iclass_rv64imafdc_branches1_u/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_branches1_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_test_all2(runner):
    """Testing aapg_iclass_rv64imafdc_test_all2.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_test_all2'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_test_all2.yaml tests/work/aapg_iclass_rv64imafdc_test_all2/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_test_all2','--output_dir=tests/work/aapg_iclass_rv64imafdc_test_all2/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_test_all2; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_user_s(runner):
    """Testing aapg_iclass_rv64imafdc_user_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_user_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_user_s.yaml tests/work/aapg_iclass_rv64imafdc_user_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_user_s','--output_dir=tests/work/aapg_iclass_rv64imafdc_user_s/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_user_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_user(runner):
    """Testing aapg_cclass_rv64imafdc_user.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_user'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_user.yaml tests/work/aapg_cclass_rv64imafdc_user/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_user','--output_dir=tests/work/aapg_cclass_rv64imafdc_user/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_user; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_bringup(runner):
    """Testing aapg_iclass_rv64imafdc_bringup.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_bringup'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_bringup.yaml tests/work/aapg_iclass_rv64imafdc_bringup/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_bringup','--output_dir=tests/work/aapg_iclass_rv64imafdc_bringup/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_bringup; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_hazards(runner):
    """Testing aapg_cclass_rv64imafdc_hazards.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_hazards'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_hazards.yaml tests/work/aapg_cclass_rv64imafdc_hazards/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_hazards','--output_dir=tests/work/aapg_cclass_rv64imafdc_hazards/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_hazards; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_user(runner):
    """Testing aapg_iclass_rv64imafdc_user.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_user'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_user.yaml tests/work/aapg_iclass_rv64imafdc_user/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_user','--output_dir=tests/work/aapg_iclass_rv64imafdc_user/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_user; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_branches1(runner):
    """Testing aapg_iclass_rv64imafdc_branches1.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_branches1'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_branches1.yaml tests/work/aapg_iclass_rv64imafdc_branches1/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_branches1','--output_dir=tests/work/aapg_iclass_rv64imafdc_branches1/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_branches1; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_exceptions_u(runner):
    """Testing aapg_cclass_rv64imafdc_exceptions_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_exceptions_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_exceptions_u.yaml tests/work/aapg_cclass_rv64imafdc_exceptions_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_exceptions_u','--output_dir=tests/work/aapg_cclass_rv64imafdc_exceptions_u/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_exceptions_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_hazards(runner):
    """Testing aapg_iclass_rv64imafdc_hazards.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_hazards'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_hazards.yaml tests/work/aapg_iclass_rv64imafdc_hazards/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_hazards','--output_dir=tests/work/aapg_iclass_rv64imafdc_hazards/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_hazards; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_recursion_branch(runner):
    """Testing aapg_iclass_rv64imafdc_recursion_branch.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_recursion_branch'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_recursion_branch.yaml tests/work/aapg_iclass_rv64imafdc_recursion_branch/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_recursion_branch','--output_dir=tests/work/aapg_iclass_rv64imafdc_recursion_branch/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_recursion_branch; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_hazards_u(runner):
    """Testing aapg_iclass_rv64imafdc_hazards_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_hazards_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_hazards_u.yaml tests/work/aapg_iclass_rv64imafdc_hazards_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_hazards_u','--output_dir=tests/work/aapg_iclass_rv64imafdc_hazards_u/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_hazards_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_exceptions_u(runner):
    """Testing aapg_iclass_rv64imafdc_exceptions_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_exceptions_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_exceptions_u.yaml tests/work/aapg_iclass_rv64imafdc_exceptions_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_exceptions_u','--output_dir=tests/work/aapg_iclass_rv64imafdc_exceptions_u/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_exceptions_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_branches1_s(runner):
    """Testing aapg_cclass_rv64imafdc_branches1_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_branches1_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_branches1_s.yaml tests/work/aapg_cclass_rv64imafdc_branches1_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_branches1_s','--output_dir=tests/work/aapg_cclass_rv64imafdc_branches1_s/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_branches1_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_bringup_u(runner):
    """Testing aapg_cclass_rv64imafdc_bringup_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_bringup_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_bringup_u.yaml tests/work/aapg_cclass_rv64imafdc_bringup_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_bringup_u','--output_dir=tests/work/aapg_cclass_rv64imafdc_bringup_u/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_bringup_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_illegal(runner):
    """Testing aapg_iclass_rv64imafdc_illegal.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_illegal'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_illegal.yaml tests/work/aapg_iclass_rv64imafdc_illegal/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_illegal','--output_dir=tests/work/aapg_iclass_rv64imafdc_illegal/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_illegal; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_bringup_s(runner):
    """Testing aapg_iclass_rv64imafdc_bringup_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_bringup_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_bringup_s.yaml tests/work/aapg_iclass_rv64imafdc_bringup_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_bringup_s','--output_dir=tests/work/aapg_iclass_rv64imafdc_bringup_s/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_bringup_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_user_s(runner):
    """Testing aapg_cclass_rv64imafdc_user_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_user_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_user_s.yaml tests/work/aapg_cclass_rv64imafdc_user_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_user_s','--output_dir=tests/work/aapg_cclass_rv64imafdc_user_s/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_user_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_illegal_u(runner):
    """Testing aapg_cclass_rv64imafdc_illegal_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_illegal_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_illegal_u.yaml tests/work/aapg_cclass_rv64imafdc_illegal_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_illegal_u','--output_dir=tests/work/aapg_cclass_rv64imafdc_illegal_u/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_illegal_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_recursion_branch(runner):
    """Testing aapg_cclass_rv64imafdc_recursion_branch.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_recursion_branch'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_recursion_branch.yaml tests/work/aapg_cclass_rv64imafdc_recursion_branch/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_recursion_branch','--output_dir=tests/work/aapg_cclass_rv64imafdc_recursion_branch/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_recursion_branch; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_bringup(runner):
    """Testing aapg_cclass_rv64imafdc_bringup.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_bringup'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_bringup.yaml tests/work/aapg_cclass_rv64imafdc_bringup/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_bringup','--output_dir=tests/work/aapg_cclass_rv64imafdc_bringup/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_bringup; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_exceptions(runner):
    """Testing aapg_iclass_rv64imafdc_exceptions.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_exceptions'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_exceptions.yaml tests/work/aapg_iclass_rv64imafdc_exceptions/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_exceptions','--output_dir=tests/work/aapg_iclass_rv64imafdc_exceptions/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_exceptions; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_branches1_s(runner):
    """Testing aapg_iclass_rv64imafdc_branches1_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_branches1_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_branches1_s.yaml tests/work/aapg_iclass_rv64imafdc_branches1_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_branches1_s','--output_dir=tests/work/aapg_iclass_rv64imafdc_branches1_s/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_branches1_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_s_recursion(runner):
    """Testing aapg_iclass_rv64imafdc_s_recursion.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_s_recursion'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_s_recursion.yaml tests/work/aapg_iclass_rv64imafdc_s_recursion/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_s_recursion','--output_dir=tests/work/aapg_iclass_rv64imafdc_s_recursion/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_s_recursion; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_illegal_u(runner):
    """Testing aapg_iclass_rv64imafdc_illegal_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_illegal_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_illegal_u.yaml tests/work/aapg_iclass_rv64imafdc_illegal_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_illegal_u','--output_dir=tests/work/aapg_iclass_rv64imafdc_illegal_u/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_illegal_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_hazards_s(runner):
    """Testing aapg_cclass_rv64imafdc_hazards_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_hazards_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_hazards_s.yaml tests/work/aapg_cclass_rv64imafdc_hazards_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_hazards_s','--output_dir=tests/work/aapg_cclass_rv64imafdc_hazards_s/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_hazards_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_user_u(runner):
    """Testing aapg_cclass_rv64imafdc_user_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_user_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_user_u.yaml tests/work/aapg_cclass_rv64imafdc_user_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_user_u','--output_dir=tests/work/aapg_cclass_rv64imafdc_user_u/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_user_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_exceptions_s(runner):
    """Testing aapg_cclass_rv64imafdc_exceptions_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_exceptions_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_exceptions_s.yaml tests/work/aapg_cclass_rv64imafdc_exceptions_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_exceptions_s','--output_dir=tests/work/aapg_cclass_rv64imafdc_exceptions_s/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_exceptions_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_test_all(runner):
    """Testing aapg_cclass_rv64imafdc_test_all.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_test_all'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_test_all.yaml tests/work/aapg_cclass_rv64imafdc_test_all/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_test_all','--output_dir=tests/work/aapg_cclass_rv64imafdc_test_all/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_test_all; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_test_all(runner):
    """Testing aapg_iclass_rv64imafdc_test_all.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_test_all'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_test_all.yaml tests/work/aapg_iclass_rv64imafdc_test_all/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_test_all','--output_dir=tests/work/aapg_iclass_rv64imafdc_test_all/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_test_all; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_recurse_low(runner):
    """Testing aapg_iclass_rv64imafdc_recurse_low.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_recurse_low'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_recurse_low.yaml tests/work/aapg_iclass_rv64imafdc_recurse_low/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_recurse_low','--output_dir=tests/work/aapg_iclass_rv64imafdc_recurse_low/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_recurse_low; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_branches1_u(runner):
    """Testing aapg_cclass_rv64imafdc_branches1_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_branches1_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_branches1_u.yaml tests/work/aapg_cclass_rv64imafdc_branches1_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_branches1_u','--output_dir=tests/work/aapg_cclass_rv64imafdc_branches1_u/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_branches1_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_exceptions(runner):
    """Testing aapg_cclass_rv64imafdc_exceptions.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_exceptions'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_exceptions.yaml tests/work/aapg_cclass_rv64imafdc_exceptions/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_exceptions','--output_dir=tests/work/aapg_cclass_rv64imafdc_exceptions/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_exceptions; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_illegal_s(runner):
    """Testing aapg_cclass_rv64imafdc_illegal_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_illegal_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_illegal_s.yaml tests/work/aapg_cclass_rv64imafdc_illegal_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_illegal_s','--output_dir=tests/work/aapg_cclass_rv64imafdc_illegal_s/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_illegal_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_recurse_med(runner):
    """Testing aapg_cclass_rv64imafdc_recurse_med.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_recurse_med'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_recurse_med.yaml tests/work/aapg_cclass_rv64imafdc_recurse_med/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_recurse_med','--output_dir=tests/work/aapg_cclass_rv64imafdc_recurse_med/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_recurse_med; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_user_u(runner):
    """Testing aapg_iclass_rv64imafdc_user_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_user_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_user_u.yaml tests/work/aapg_iclass_rv64imafdc_user_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_user_u','--output_dir=tests/work/aapg_iclass_rv64imafdc_user_u/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_user_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_bringup_u(runner):
    """Testing aapg_iclass_rv64imafdc_bringup_u.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_bringup_u'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_bringup_u.yaml tests/work/aapg_iclass_rv64imafdc_bringup_u/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_bringup_u','--output_dir=tests/work/aapg_iclass_rv64imafdc_bringup_u/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_bringup_u; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_illegal(runner):
    """Testing aapg_cclass_rv64imafdc_illegal.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_illegal'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_illegal.yaml tests/work/aapg_cclass_rv64imafdc_illegal/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_illegal','--output_dir=tests/work/aapg_cclass_rv64imafdc_illegal/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_illegal; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_cclass_rv64imafdc_recurse_low(runner):
    """Testing aapg_cclass_rv64imafdc_recurse_low.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_cclass_rv64imafdc_recurse_low'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_cclass_rv64imafdc_recurse_low.yaml tests/work/aapg_cclass_rv64imafdc_recurse_low/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_cclass_rv64imafdc_recurse_low','--output_dir=tests/work/aapg_cclass_rv64imafdc_recurse_low/asm'])
    try:
        os.system('cd tests/work/aapg_cclass_rv64imafdc_recurse_low; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_exceptions_s(runner):
    """Testing aapg_iclass_rv64imafdc_exceptions_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_exceptions_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_exceptions_s.yaml tests/work/aapg_iclass_rv64imafdc_exceptions_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_exceptions_s','--output_dir=tests/work/aapg_iclass_rv64imafdc_exceptions_s/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_exceptions_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_illegal_s(runner):
    """Testing aapg_iclass_rv64imafdc_illegal_s.yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_illegal_s'])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_illegal_s.yaml tests/work/aapg_iclass_rv64imafdc_illegal_s/config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_illegal_s','--output_dir=tests/work/aapg_iclass_rv64imafdc_illegal_s/asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_illegal_s; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0

@pytest.mark.timeout(300)
def test_aapg_iclass_rv64imafdc_data_trashing (runner):
    """Testing aapg_iclass_rv64imafdc_data_trashing .yaml config file in parallel mode"""
    make = 0
    runner.invoke(cli, ['setup','--setup_dir=tests/work/aapg_iclass_rv64imafdc_data_trashing '])
    os.system('cp tests/ci_cd_templates/core_configs/aapg_iclass_rv64imafdc_data_trashing .yaml tests/work/aapg_iclass_rv64imafdc_data_trashing /config.yaml')
    result = runner.invoke(cli, ['gen','--setup_dir=tests/work/aapg_iclass_rv64imafdc_data_trashing ','--output_dir=tests/work/aapg_iclass_rv64imafdc_data_trashing /asm'])
    try:
        os.system('cd tests/work/aapg_iclass_rv64imafdc_data_trashing ; make')
    except:
        make = 1
    assert result.exit_code == 0 and make == 0









crt_asm = '''
#include "encoding.h"

#if __riscv_xlen == 64
  # define LREG ld
  # define LREGU lwu
  # define SREG sd
  # define REGBYTES 8
  # define FMV fmv.d.x
#else
  # define LREG lw
  # define SREG sw
  # define LREGU lw
  # define REGBYTES 4
  # define FMV fmv.w.x
#endif

  .section ".text.init"
  .globl _start
_start:
  la sp, <!data_section!>
  LREG  x1, 0*REGBYTES(sp)
  LREG  x3, 1*REGBYTES(sp)
  LREG  x4, 2*REGBYTES(sp)
  LREG  x5, 3*REGBYTES(sp)
  LREG  x6, 4*REGBYTES(sp)
  LREG  x7, 5*REGBYTES(sp)
  LREG  x8, 6*REGBYTES(sp)
  LREG  x9, 7*REGBYTES(sp)
  LREG  x10,8*REGBYTES(sp)
  LREG  x11,9*REGBYTES(sp)
  LREG  x12,10*REGBYTES(sp)
  LREG  x13,11*REGBYTES(sp)
  LREG  x14,12*REGBYTES(sp)
  LREG  x15,13*REGBYTES(sp)
  LREG  x16,14*REGBYTES(sp)
  LREG  x17,15*REGBYTES(sp)
  LREG  x18,16*REGBYTES(sp)
  LREG  x19,17*REGBYTES(sp)
  LREG  x20,18*REGBYTES(sp)
  LREG  x21,19*REGBYTES(sp)
  LREG  x22,20*REGBYTES(sp)
  LREG  x23,21*REGBYTES(sp)
  LREG  x24,22*REGBYTES(sp)
  LREG  x25,23*REGBYTES(sp)
  LREG  x26,24*REGBYTES(sp)
  LREG  x27,25*REGBYTES(sp)
  LREG  x28,26*REGBYTES(sp)
  LREG  x29,27*REGBYTES(sp)
  LREG  x30,28*REGBYTES(sp)
  LREG  x31,29*REGBYTES(sp)

  # enable FPU and accelerator if present
  li t0, MSTATUS_FS | MSTATUS_XS
  csrs mstatus, t0

  # make sure XLEN agrees with compilation choice
  li t0, 1
  slli t0, t0, 31
#if __riscv_xlen == 64
  bgez t0, 1f
#else
  bltz t0, 1f
#endif
2:
  li a0, 1
  sw a0, tohost, t0
  j 2b
1:

#ifdef __riscv_flen
  # initialize FPU if we have one
  la t0, 1f
  csrw mtvec, t0

  fssr    x0
  FMV  f0, x1
  FMV  f1, x1
  FMV  f2, x2
  FMV  f3, x3
  FMV  f4, x4
  FMV  f5, x5
  FMV  f6, x6
  FMV  f7, x7
  FMV  f8, x8
  FMV  f9, x9
  FMV  f10, x10
  FMV  f11, x11
  FMV  f12, x12
  FMV  f13, x13
  FMV  f14, x14
  FMV  f15, x15
  FMV  f16, x16
  FMV  f17, x17
  FMV  f18, x18
  FMV  f19, x19
  FMV  f20, x20
  FMV  f21, x21
  FMV  f22, x22
  FMV  f23, x23
  FMV  f24, x24
  FMV  f25, x25
  FMV  f26, x26
  FMV  f27, x27
  FMV  f28, x28
  FMV  f29, x29
  FMV  f30, x30
  FMV  f31, x31
1:
#endif

  # initialize trap vector
  la t0, trap_entry
  csrw mtvec, t0

  la  tp, _end + 63
  and tp, tp, -64

  # get core id
  csrr a0, mhartid
  # for now, assume only 1 core
  li a1, 1
1:bgeu a0, a1, 1b

  # give each core 128KB of stack + TLS
#define STKSHIFT 17
  sll a2, a0, STKSHIFT
  add tp, tp, a2
  add sp, a0, 1
  sll sp, sp, STKSHIFT
  add sp, sp, tp

  j main

  .align 2
trap_entry:
  addi sp, sp, -272

  SREG x1, 1*REGBYTES(sp)
  SREG x2, 2*REGBYTES(sp)
  SREG x3, 3*REGBYTES(sp)
  SREG x4, 4*REGBYTES(sp)
  SREG x5, 5*REGBYTES(sp)
  SREG x6, 6*REGBYTES(sp)
  SREG x7, 7*REGBYTES(sp)
  SREG x8, 8*REGBYTES(sp)
  SREG x9, 9*REGBYTES(sp)
  SREG x10, 10*REGBYTES(sp)
  SREG x11, 11*REGBYTES(sp)
  SREG x12, 12*REGBYTES(sp)
  SREG x13, 13*REGBYTES(sp)
  SREG x14, 14*REGBYTES(sp)
  SREG x15, 15*REGBYTES(sp)
  SREG x16, 16*REGBYTES(sp)
  SREG x17, 17*REGBYTES(sp)
  SREG x18, 18*REGBYTES(sp)
  SREG x19, 19*REGBYTES(sp)
  SREG x20, 20*REGBYTES(sp)
  SREG x21, 21*REGBYTES(sp)
  SREG x22, 22*REGBYTES(sp)
  SREG x23, 23*REGBYTES(sp)
  SREG x24, 24*REGBYTES(sp)
  SREG x25, 25*REGBYTES(sp)
  SREG x26, 26*REGBYTES(sp)
  SREG x27, 27*REGBYTES(sp)
  SREG x28, 28*REGBYTES(sp)
  SREG x29, 29*REGBYTES(sp)
  SREG x30, 30*REGBYTES(sp)
  SREG x31, 31*REGBYTES(sp)

  csrr a0, mcause                 # copy the mcause to register a0.
  csrr a1, mepc                   # copy the mepc to register a1.
  lhu  a2, 0(a1)                  # load instruction into reg a1.

  # check the lower 2 bits to see if the instruction is 32-bit or 16-bit.
  andi a2, a2, 0x3;
  li t0, 0x3
  bne a2,t0,inst16

  inst32:                           # is 32-bit instruction then increment by 4
  addi a1,a1,0x4
  beqz x0,1f

inst16:
  addi a1,a1,0x2                  # is 16-bit instruction then increment by 2

1: 
  csrw mepc, a1                   # point mepc to the next instruction.

  # use mcause to update the number of exceptions encountered in the program.

  # Remain in M-mode after eret
  li t0, MSTATUS_MPP
  csrs mstatus, t0

  LREG x1, 1*REGBYTES(sp)
  LREG x2, 2*REGBYTES(sp)
  LREG x3, 3*REGBYTES(sp)
  LREG x4, 4*REGBYTES(sp)
  LREG x5, 5*REGBYTES(sp)
  LREG x6, 6*REGBYTES(sp)
  LREG x7, 7*REGBYTES(sp)
  LREG x8, 8*REGBYTES(sp)
  LREG x9, 9*REGBYTES(sp)
  LREG x10, 10*REGBYTES(sp)
  LREG x11, 11*REGBYTES(sp)
  LREG x12, 12*REGBYTES(sp)
  LREG x13, 13*REGBYTES(sp)
  LREG x14, 14*REGBYTES(sp)
  LREG x15, 15*REGBYTES(sp)
  LREG x16, 16*REGBYTES(sp)
  LREG x17, 17*REGBYTES(sp)
  LREG x18, 18*REGBYTES(sp)
  LREG x19, 19*REGBYTES(sp)
  LREG x20, 20*REGBYTES(sp)
  LREG x21, 21*REGBYTES(sp)
  LREG x22, 22*REGBYTES(sp)
  LREG x23, 23*REGBYTES(sp)
  LREG x24, 24*REGBYTES(sp)
  LREG x25, 25*REGBYTES(sp)
  LREG x26, 26*REGBYTES(sp)
  LREG x27, 27*REGBYTES(sp)
  LREG x28, 28*REGBYTES(sp)
  LREG x29, 29*REGBYTES(sp)
  LREG x30, 30*REGBYTES(sp)
  LREG x31, 31*REGBYTES(sp)

  addi sp, sp, 272
  mret

.section ".tdata.begin"
.globl _tdata_begin
_tdata_begin:

.section ".tdata.end"
.globl _tdata_end
_tdata_end:

.section ".tbss.end"
.globl _tbss_end
_tbss_end:

.section ".tohost","aw",@progbits
.align 6
.globl tohost
tohost: .dword 0
.align 6
.globl fromhost
fromhost: .dword 0

# Exception generation numbers
.align 4
.globl ecause_num
ecause_num:
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0
    .dword 0

# Read only data
.section ".rodata"
.globl rodata
rodata:
    <!rodata_config!>
'''
